body {
    padding-top: 20px;
    padding-bottom: 40px;
}

/* Custom container */
.container-narrow {
    margin: 0 auto;
    max-width: 675px;
}
.container-narrow > hr {
    margin: 30px 0;
}

/* Main marketing message and sign up button */
.jumbotron {
    margin: 60px 0;
    text-align: center;
}
.jumbotron h1 {
    font-size: 72px;
    line-height: 1;
}
.jumbotron .btn {
    font-size: 21px;
    padding: 14px 24px;
}

/* Supporting marketing content */
.marketing {
    margin: 60px 0;
}
.marketing p + h4 {
    margin-top: 28px;
}

.captcha {
    margin: auto;
    display: block;
    width: 250px;
}

.captcha .btn {
    margin: auto;
    width: 100px;
    display: block;
}

.fixed-size-btn {
    margin: 10px;
    width: 200px;
}

.main-steps {
    margin: 50px;
}


.main-btns {
    margin: auto;
    width: 450px;
    display: block;
}

.step{
border: 1px solid #ccc;
border: 1px solid rgba(0, 0, 0, 0.2);
-webkit-border-radius: 6px;
-moz-border-radius: 6px;
border-radius: 6px;
-webkit-box-shadow: 0 5px 10px rgba(0, 0, 0, 0.2);
-moz-box-shadow: 0 5px 10px rgba(0, 0, 0, 0.2);
box-shadow: 0 5px 10px rgba(0, 0, 0, 0.2);
-webkit-background-clip: padding-box;
-moz-background-clip: padding;
background-clip: padding-box;
padding: 10px 20px 0px;
margin-bottom: 10px;
}

.step-title {
color: #808080;
font-size: 18px;
font-weight: 100;
}

.step-text {
    font-size: 18px;
line-height: 30px;
margin-top: 2px;
}
.lead_right {
    margin-bottom: 20px;
    font-size: 21px;
    font-weight: 200;
    line-height: 30px
}
[class*="bdb_span"] {
    min-height: 1px;
    margin: 0 20px 16px 20px;
}
.bdb_span7 {
    width: 560px
}

div.bridge-lines {
    padding: 20px;
    margin: 0px 0px 20px;
    min-height: 20px;
    font-family: Monaco,Menlo,Consolas,"Courier New",monospace;
    font-size: 95%;
    line-height: 175%;
    color: rgb(44, 62, 80);
    word-break: break-all;
    word-wrap: normal;
    white-space: nowrap;
    overflow-x: auto;
    z-index: 1000;
    background-color: rgb(236, 240, 241);
    border: 0px solid transparent;
    border-radius: 6px 6px 6px 6px;
    border-color: #2C3E50;
    box-shadow: none;
    box-sizing: border-box;
    -moz-box-sizing: border-box;
    cursor: copy;
}

div.bridge-lines.-webkit-scrollbar {
    width: 9px;
    height: 9px;
}
div.bridge-lines.-webkit-scrollbar-button.start.decrement {
    display: block;
    height: 0;
    background-color: transparent;
}
div.bridge-lines.-webkit-scrollbar-button.end.increment {
    display: block;
    height: 0;
    background-color: transparent;
}
div.bridge-lines.-webkit-scrollbar-track-piece {
    background-color: #FAFAFA;
    -webkit-border-radius: 0;
    -webkit-border-bottom-right-radius: 6px;
    -webkit-border-bottom-left-radius: 6px;
}
div.bridge-lines.-webkit-scrollbar-thumb.vertical {
    height: 50px;
    background-color: #999;
    -webkit-border-radius: 6px;
}
div.bridge-lines.-webkit-scrollbar-thumb.horizontal{
    width: 50px;
    background-color: #999;
    -webkit-border-radius: 6px;
}
